# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: MsgLib.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cMsgLib.proto\x1a\x19google/protobuf/any.proto\"A\n\x0bmsgEnvelope\x12\x0f\n\x07msgName\x18\x01 \x01(\t\x12!\n\x03msg\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Anyb\x06proto3')



_MSGENVELOPE = DESCRIPTOR.message_types_by_name['msgEnvelope']
msgEnvelope = _reflection.GeneratedProtocolMessageType('msgEnvelope', (_message.Message,), {
  'DESCRIPTOR' : _MSGENVELOPE,
  '__module__' : 'MsgLib_pb2'
  # @@protoc_insertion_point(class_scope:msgEnvelope)
  })
_sym_db.RegisterMessage(msgEnvelope)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MSGENVELOPE._serialized_start=43
  _MSGENVELOPE._serialized_end=108
# @@protoc_insertion_point(module_scope)



import logging
import google.protobuf.symbol_database
import google.protobuf.descriptor_pool
import google.protobuf.message_factory
#from google.protobuf.any_pb2 import Any
from dividere import MsgLib
from dividere import connection

#================================================================================
#-- Encoder/Decoder class; takes in protobuf message, encloses it in a envelope
#--  message for transport and allowd decoding from the received message
#--  primarily used in conjunction with transport classes in this package
#================================================================================
class ProtoBuffEncoder:
  '''
    This class suports taking in a user protobuf message and encode/pack
    into a container message for transport.  This is one end of a encode/decode
    sequence used when sending a user message through a socket while allowing
    a variety of messages to be sent thru a shared socket channel.
    This is one end of the encode/decode sequence; encoding done at the sending
    end, decoding at the receiving end.
  '''
  def __init__(self):
    '''
      Initialize object resources
    '''
    pass

  def encode(self, msg):
    '''
      Encapsulate the specified message into a container message for
      transport and return it to the caller
    '''
    env=MsgLib.msgEnvelope()
    env.msgName=msg.__class__.__name__
    env.msg.Pack(msg)
    return env

class ProtoBuffDecoder:
  '''
    This class suports taking in a user protobuf message and encode/pack
    into a container message for transport.  This is one end of a encode/decode
    sequence used when sending a user message through a socket while allowing
    a variety of messages to be sent thru a shared socket channel.
    This is one end of the encode/decode sequence; encoding done at the sending
    end, decoding at the receiving end.
  '''
  def __init__(self):
    pass

  def decode(self, msgEnv):
    '''
      Extract the user message from the specified container message
      and return it to the caller.
    '''
    msgDesc=google.protobuf.descriptor_pool.Default().FindMessageTypeByName(msgEnv.msgName)
    factory=google.protobuf.message_factory.MessageFactory()
    msgClass=factory.GetPrototype(msgDesc)
    c=msgClass()
    msgEnv.msg.Unpack(c)
    return c

class Publisher:
  '''
    Similar functionality to the Publish/Subscriber pairing in the connection
    module, differing in the expected user message being sent.  The messaging
    module specializes in sending/receiving protobuf-based messages.
  '''
  def __init__(self,endPoint):
    '''
      Create a publisher connection and encoder
    '''
    #--create pub component and encoder
    self.pub_=connection.Publisher(endPoint)
    self.encoder_=ProtoBuffEncoder()

  def __del__(self):
    '''
      Free allocated object resources
    '''
    self.pub_=None
    self.encoder_=None

  def send(self, msg):
    '''
      Encode message into envelope container, convert it to
      a byte stream and send out wire via the connector
    '''
    env=self.encoder_.encode(msg)
    self.pub_.send(env.SerializeToString())

class Subscriber:
  '''
    Similar functionality to the Publish/Subscriber pairing in the connection
    module, differing in the expected user message being sent.  The messaging
    module specializes in sending/receiving protobuf-based messages.
  '''
  @staticmethod
  def topicId(msg):
    '''
      Translate a protobuf message into a topic name
      (the beginning of the string coming across the 'wire')
      used to subscribe to specific message(s)
      Note: expected usage is internal to the module, not
      intended for external use
    '''
    return '\n\x08%s'%(msg.__class__.__name__)

  def __init__(self,endPoint, msgSubList=[]):
    '''
       Allocate all necessary resources, subscribe to messages.
       If message subscription list is empty, subscribe to all messages
       otherwise subscribe to the specified messages exclusively
       create subscriber object and decoder components
    '''
    if (len(msgSubList)==0):
      topic=''
    else:
      topic=self.topicId(msgSubList[0])
    self.sub_=connection.Subscriber(endPoint, topic)
    self.decoder_=ProtoBuffDecoder()
    for topicMsg in msgSubList[1:]:
      self.sub_.subscribe(self.topicId(topicMsg))

  def __del__(self):
    '''
      Free all allocated object resources
    '''
    self.sub_=None
    self.decoder_=None

  def recv(self):
    '''
      Retrieve byte stream from subscriber, parse byte stream into envelope
       message, then decode and return the contained user message
    '''
    S=self.sub_.recv()
    env=MsgLib.msgEnvelope()
    env.ParseFromString(S)
    return self.decoder_.decode(env)

class Request:
  '''
    Similar functionality to the Request/Response pairing in the connection
    module, differing in the expected user message being sent.  The messaging
    module specializes in sending/receiving protobuf-based messages.
  '''
  def __init__(self,endPoint):
    '''
      Create a request connection and encoder
    '''
    #--create req component and encoder
    self.sock_=connection.Request(endPoint)
    self.encoder_=ProtoBuffEncoder()
    self.decoder_=ProtoBuffDecoder()

  def __del__(self):
    '''
      Free allocated object resources
    '''
    self.sock_=None
    self.encoder_=None

  def send(self, msg):
    '''
      Encode message into envelope container, convert it to
      a byte stream and send out wire via the connector
    '''
    env=self.encoder_.encode(msg)
    self.sock_.send(env.SerializeToString())

  def recv(self):
    '''
      Retrieve byte stream from repscriber, parse byte stream into envelope
       message, then decode and return the contained user message
    '''
    S=self.sock_.recv()
    env=MsgLib.msgEnvelope()
    env.ParseFromString(S)
    return self.decoder_.decode(env)

class Response:
  '''
    Similar functionality to the Request/Response pairing in the connection
    module, differing in the expected user message being sent.  The messaging
    module specializes in sending/receiving protobuf-based messages.
  '''

  def __init__(self,endPoint):
    '''
       Allocate all necessary resources, repscribe to messages.
       If message repscription list is empty, repscribe to all messages
       otherwise repscribe to the specified messages exclusively
       create repscriber object and decoder components
    '''
    self.sock_=connection.Response(endPoint)
    self.decoder_=ProtoBuffDecoder()
    self.encoder_=ProtoBuffEncoder()

  def __del__(self):
    '''
      Free all allocated object resources
    '''
    self.sock_=None
    self.decoder_=None

  def recv(self):
    '''
      Retrieve byte stream from repscriber, parse byte stream into envelope
       message, then decode and return the contained user message
    '''
    S=self.sock_.recv()
    env=MsgLib.msgEnvelope()
    env.ParseFromString(S)
    return self.decoder_.decode(env)

  def send(self, msg):
    '''
      Encode message into envelope container, convert it to
      a byte stream and send out wire via the connector
    '''
    env=self.encoder_.encode(msg)
    self.sock_.send(env.SerializeToString())




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '1.1.0'
DESCRIPTION = 'Distributed Systems Framework'
LONG_DESCRIPTION = 'Dīvidere, latin for "to divide, to seperate" seemed an appropriate package name for a distributed system framework project. \n\nThe primary goal of dīvidere is to combine three complementary technologies (Python, ZeroMQ, Protobuf) into a distributed system messaging framework.  ZeroMQ will provide a variety of transport mechanisms, Protobuf providing a language-independent, strongly-typed message encoding and Python the means to combine these components into a reusable framework.'

# Setting up
setup(
    name="dividere",
    version=VERSION,
    author="Grant Lipelt",
    author_email="<lipeltgm@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['zmq','protobuf>=3.19.0'],
    keywords=['python', 'ZeroMq', 'zmq', '0mq', 'protobuf'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
#       "Operating System :: MacOS :: MacOS X",
#       "Operating System :: Microsoft :: Windows"
    ]
)



import logging
import zmq
import time
import threading
from zmq.utils.monitor import recv_monitor_message

class Connector:
  '''
     This abstract class defines the interfaces and structures
     for ZMQ socket-based derived classes.  This class provides
     the ZMQ context and socket event monitoring useful for
     debugging socket state changes.
     The socket monitoring is conducted by an independent thread,
     which is terminated/joined at object termination.
  '''
  @staticmethod
  def socketEventMonitor(monitorSock):
    '''
      Background threading callback, supports monitoring the
      specified socket via a background thread and logs state
      changes of the socket for debugging purposes.
      Monitors the socket until monitoring is terminated
      via object destructor (e.g. obj = None)
      Note: Used internally to class(es), not intended for external usage
    '''
    EVENT_MAP = {}
    for name in dir(zmq):
      if name.startswith('EVENT_'):
        value = getattr(zmq, name)
        EVENT_MAP[value] = name

    while monitorSock.poll():
      evt: Dict[str, Any] = {}
      mon_evt = recv_monitor_message(monitorSock)
      evt.update(mon_evt)
      evt['description'] = EVENT_MAP[evt['event']]
      logging.debug(f"Event: {evt}")
      if evt['event'] == zmq.EVENT_MONITOR_STOPPED:
        break

    monitorSock.close()

  @staticmethod
  def registerSocketMonitoring(sock):
    '''
    Creates a monitoring thread for the specified socket,
    starts the thread and returns the thread id to the caller
    which allows joining on the thread post stopping monitoring
    Note: Used internally to class(es), not intended for external usage
    '''
    monitorSock = sock.get_monitor_socket()
    tid = threading.Thread(target=Connector.socketEventMonitor, args=(monitorSock,))
    tid.start()
    return tid;

  def __init__(self):
    '''
      Creates resources used in base classes and defines expected
      structure to be used in derived classes. 
    '''
    self.ctx_=zmq.Context()
    self.socket_ = None
    self.tid_ = None

  def __del__(self):
    '''
    Performs cleanup for all allocated resources;
    disable monitoring, wait for monitoring thread completes,
    close the socket and close the context
    '''
    self.socket_.disable_monitor()
    if self.tid_: self.tid_.join() 
    self.socket_.close()
    self.ctx_.term()

#================================================================================
#  Pub/Sub Connection Pair
#================================================================================
class Publisher(Connector):
  '''
     This class creates a publisher socket at the specified endpoint.
     This is the pub in the Pub/Sub pattern.
  '''
  def __init__(self, endPoint):
    '''
       Allocate base class resources, create PUB socket, start
       socket debug monitoring and connect the socket to the
       specified endpoint (e.g. 'tcp://*:5555')
       Refer to ZMQ documentation for details on available transport
       and syntax of endpoint.
    '''
    super(self.__class__,self).__init__()
    self.socket_=self.ctx_.socket(zmq.PUB)
    self.tid_=self.registerSocketMonitoring(self.socket_)
    self.socket_.bind(endPoint)

  def send(self, msg):
    '''
      Publish the specified message (expected sequence of bytes)
    '''
    self.socket_.send(msg)

class Subscriber(Connector):
  '''
     This class creates a subscriber socket at the specified endpoint.
     This is the sub in the Pub/Sub pattern.  By default, a subscriber
     object will listen for all messages, but can be filtered by specifying
     a topic(s); either by specifying a topic during the initializer or
     calling subscribe() after object creation
  '''
  def __init__(self, endPoint, topic=''):
    '''
       Allocate base class resources, create SUB socket, start
       socket debug monitoring and connect the socket to the
       specified endpoint (e.g. 'tcp://localhost:5555')
       Subscribes to the specified topic, by default the object
       will receive all messages.
       Refer to ZMQ documentation for details on available transport
       and syntax of endpoint.
    '''
    super(self.__class__,self).__init__()
    self.socket_=self.ctx_.socket(zmq.SUB)
    self.tid_=self.registerSocketMonitoring(self.socket_)
    self.socket_.connect(endPoint)
    self.subscribe(topic)

  def subscribe(self, topic):
    '''
      Allows subscribing to additional topics (beyond the one
      specified in the constructor)
    '''
    self.socket_.setsockopt_string(zmq.SUBSCRIBE, topic)

  def recv(self):
    '''
      Wait for next message to arrive and return it to the
      caller.
    '''
    S=self.socket_.recv()
    return S

#================================================================================
#  Request/Response Connection Pair
#================================================================================
class Request(Connector):
  '''
    First part of a Request/Response connection pair.  Request object
    initiates all messages, response object sends message response.
    Failure to adhere to this sender protocol will result in exception
    being thrown.
    Note: this pairing allows for 1-N cardinality, one request connection
          object sending to N-response objects.  When configured like this
          the recipient of any message is routed in a round-robin fashion
          to one response object
  '''
  def __init__(self, endPointList):
    '''
      Allocate all resources to support the object;         
      create a socket, register it for monitoring, and connect
      it to the specified endpoint
    '''
    super(self.__class__,self).__init__()
    self.socket_=self.ctx_.socket(zmq.REQ)
    self.tid_=self.registerSocketMonitoring(self.socket_)
    for endPt in endPointList:
      logging.debug("binding to %s"%(endPt))
      self.socket_.connect(endPt)

  def send(self, msg):
    '''
      Send the specified message out the socket channel.  
      Message consists of a stream of bytes.
    '''
    S=self.socket_.send(msg)

  def recv(self):
    '''
      Wait for and return the incoming message.
    '''
    S=self.socket_.recv()
    return S

class Response(Connector):
  '''
    First part of a Request/Response connection pair.  Request object
    initiates all messages, response object sends message response.
    Failure to adhere to this sender protocol will result in exception
    being thrown.
  '''
  def __init__(self, endPoint):
    '''
      Allocate all resources to support the object;         
      create a socket, register it for monitoring, and connect
      it to the specified endpoint
    '''
    super(self.__class__,self).__init__()
    self.socket_=self.ctx_.socket(zmq.REP)
    self.tid_=self.registerSocketMonitoring(self.socket_)
    logging.debug("binding to %s"%(endPoint))
    self.socket_.bind(endPoint)

  def send(self, msg):
    '''
      Send the specified message out the socket channel
      Message consists of a stream of bytes.
    '''
    S=self.socket_.send(msg)

  def recv(self):
    '''
      Wait for and return the incoming message.
    '''
    S=self.socket_.recv()
    return S

